Public Delegate Function MyFxDelegate(ByVal nNumVars As Integer, ByRef fX() As Double, ByRef fParam() As Double) As Double
Public Delegate Function SayFxDelegate() As String

Public Class COptimNewton1
  Dim m_MyFx As MyFxDelegate

  Protected Function MyFxEx(ByVal nNumVars As Integer, _
              ByRef fX() As Double, ByRef fParam() As Double, _
              ByRef fDeltaX() As Double, ByVal fLambda As Double) As Double
    Dim I As Integer
    Dim fXX(nNumVars) As Double

    For I = 0 To nNumVars - 1
      fXX(I) = fX(I) + fLambda * fDeltaX(I)
    Next I

    MyFxEx = m_MyFx(nNumVars, fXX, fParam)
  End Function

  Protected Function FirstDeriv(ByVal nNumVars As Integer, _
               ByRef fX() As Double, ByRef fParam() As Double, _
               ByVal nIdxI As Integer) As Double
    Dim fXt, h, Fp, Fm As Double

    fXt = fX(nIdxI)
    h = 0.01 * (1 + Math.Abs(fXt))
    fX(nIdxI) = fXt + h
    Fp = m_MyFx(nNumVars, fX, fParam)
    fX(nIdxI) = fXt - h
    Fm = m_MyFx(nNumVars, fX, fParam)
    fX(nIdxI) = fXt
    FirstDeriv = (Fp - Fm) / 2 / h
  End Function

  Protected Function SecondDeriv(ByVal nNumVars As Integer, _
                 ByRef fX() As Double, ByRef fParam() As Double, _
                 ByVal nIdxI As Integer, ByVal nIdxJ As Integer) As Double
    Dim fXt, fYt, fHX, fHY, F0, Fp, Fm As Double
    Dim Fpp, Fmm, Fpm, Fmp, fResult As Double

    ' calculate second derivative?
    If nIdxI = nIdxJ Then
      F0 = m_MyFx(nNumVars, fX, fParam)
      fXt = fX(nIdxI)
      fHX = 0.01 * (1 + Math.Abs(fXt))
      fX(nIdxI) = fXt + fHX
      Fp = m_MyFx(nNumVars, fX, fParam)
      fX(nIdxI) = fXt - fHX
      Fm = m_MyFx(nNumVars, fX, fParam)
      fX(nIdxI) = fXt
      fResult = (Fp - 2 * F0 + Fm) / fHX ^ 2
    Else
      fXt = fX(nIdxI)
      fYt = fX(nIdxJ)
      fHX = 0.01 * (1 + Math.Abs(fXt))
      fHY = 0.01 * (1 + Math.Abs(fYt))
      ' calculate Fpp
      fX(nIdxI) = fXt + fHX
      fX(nIdxJ) = fYt + fHY
      Fpp = m_MyFx(nNumVars, fX, fParam)
      ' calculate Fmm
      fX(nIdxI) = fXt - fHX
      fX(nIdxJ) = fYt - fHY
      Fmm = m_MyFx(nNumVars, fX, fParam)
      ' calculate Fpm
      fX(nIdxI) = fXt + fHX
      fX(nIdxJ) = fYt - fHY
      Fpm = m_MyFx(nNumVars, fX, fParam)
      ' calculate Fmp
      fX(nIdxI) = fXt - fHX
      fX(nIdxJ) = fYt + fHY
      Fmp = m_MyFx(nNumVars, fX, fParam)
      fX(nIdxI) = fXt
      fX(nIdxJ) = fYt
      fResult = (Fpp - Fmp - Fpm + Fmm) / (4 * fHX * fHY)
    End If

    Return fResult
  End Function


  Protected Sub GetFirstDerives(ByVal nNumVars As Integer, _
                  ByRef fX() As Double, ByRef fParam() As Double, _
                  ByRef fFirstDerivX() As Double)

    Dim I As Integer

    For I = 0 To nNumVars - 1
      fFirstDerivX(I) = FirstDeriv(nNumVars, fX, fParam, I)
    Next I
  End Sub

  Protected Sub GetSecondDerives(ByVal nNumVars As Integer, _
                  ByRef fX() As Double, ByRef fParam() As Double, _
                  ByRef fSecondDerivX(,) As Double)

    Dim I, J As Integer

    For I = 0 To nNumVars - 1
      For J = 0 To nNumVars - 1
        fSecondDerivX(I, J) = SecondDeriv(nNumVars, fX, fParam, I, J)
      Next J
    Next I
  End Sub

  Public Function CalcOptim(ByVal nNumVars As Integer, ByRef fX() As Double, ByRef fParam() As Double, _
                    ByRef fToler() As Double, ByVal fEpsFx As Double, ByVal nMaxIter As Integer, _
                    ByRef nIter As Integer, ByRef sErrorMsg As String, _
                    ByVal MyFx As MyFxDelegate) As Double

    Dim I As Integer
    Dim fNorm As Double
    Dim g(nNumVars) As Double, Index(nNumVars) As Integer
    Dim DeltaX(nNumVars) As Double, J(nNumVars, nNumVars) As Double
    Dim bStop As Boolean

    m_MyFx = MyFx

    nIter = 1
    Do
      nIter += 1
      If nIter > nMaxIter Then
        sErrorMsg = "Reached maximum iterations limit"
        Exit Do
      End If

      GetFirstDerives(nNumVars, fX, fParam, g)

      ' test if gradient is shallow enough
      fNorm = MatrixLibVb.VectNorm(g)
      If fNorm < fEpsFx Then Exit Do

      GetSecondDerives(nNumVars, fX, fParam, J)
      MatrixLibVb.MV_LUDecomp(J, Index, nNumVars)
      MatrixLibVb.MV_LUBackSubst(J, Index, nNumVars, g)
      For I = 0 To nNumVars - 1
        DeltaX(I) = g(I)
        fX(I) -= DeltaX(I)
      Next I

      bStop = True
      For I = 0 To nNumVars - 1
        If Math.Abs(DeltaX(I)) > fToler(I) Then
          bStop = False
          Exit For
        End If
      Next I

    Loop Until bStop

    Return MyFx(nNumVars, fX, fParam)
  End Function
End Class
